=== modified file 'NEWS'
--- NEWS	2007-07-04 03:12:37 +0000
+++ NEWS	2007-07-04 08:08:08 +0000
@@ -103,6 +103,16 @@
       These changes are an API break in the testing infrastructure only.
       (Robert Collins)
 
+    * ``bzrlib.add.smart_add_tree`` will no longer perform glob expansion on
+      win32. Callers of the function should do this and use the new
+      ``MutableTree.smart_add`` method instead. (Robert Collins)
+
+    * ``bzrlib.add.glob_expand_for_win32`` is now
+      ``bzrlib.win32utils.glob_expand``.  (Robert Collins)
+
+    * ``bzrlib.add.FastPath`` is now private and moved to 
+      ``bzrlib.mutabletree._FastPath``. (Robert Collins, Martin Pool)
+
   INTERNALS:
 
     * New SMTPConnection class to unify email handling.  (Adeodato Simó)
@@ -111,6 +121,14 @@
       allows a nicer UI when hooks are running as the current hook can
       be displayed. (Robert Collins)
 
+    * New method ``_maybe_expand_globs`` on the ``Command`` class for 
+      dealing with unexpanded glob lists - e.g. on the win32 platform. This
+      was moved from ``bzrlib.add._prepare_file_list``. (Robert Collins)
+
+    * ``bzrlib.add.smart_add`` and ``bzrlib.add.smart_add_tree`` are now
+      deprecated in favour of ``MutableTree.smart_add``. (Robert Collins,
+      Martin Pool)
+
   TESTING:
 
     * Removed the ``--keep-output`` option from selftest and clean up test

=== modified file 'bzrlib/add.py'
--- bzrlib/add.py	2007-04-09 21:06:31 +0000
+++ bzrlib/add.py	2007-07-03 02:33:32 +0000
@@ -18,53 +18,15 @@
 
 import errno
 import os
-from os.path import dirname
 import sys
 
 import bzrlib.bzrdir
 import bzrlib.errors as errors
-from bzrlib.inventory import InventoryEntry
-from bzrlib.trace import mutter, note, warning
-from bzrlib.errors import NotBranchError
 import bzrlib.osutils
+from bzrlib.symbol_versioning import *
 from bzrlib.workingtree import WorkingTree
 
 
-def glob_expand_for_win32(file_list):
-    """Replacement for glob expansion by the shell.
-
-    Win32's cmd.exe does not do glob expansion (eg ``*.py``), so we do our own
-    here.
-
-    :param file_list: A list of filenames which may include shell globs.
-    :return: An expanded list of filenames.
-    """
-    if not file_list:
-        return
-    import glob
-    expanded_file_list = []
-    for possible_glob in file_list:
-        glob_files = glob.glob(possible_glob)
-
-        if glob_files == []:
-            # special case to let the normal code path handle
-            # files that do not exists
-            expanded_file_list.append(possible_glob)
-        else:
-            expanded_file_list += glob_files
-    return expanded_file_list
-
-
-def _prepare_file_list(file_list):
-    """Prepare a file list for use by smart_add_*."""
-    if sys.platform == 'win32':
-        file_list = glob_expand_for_win32(file_list)
-    if not file_list:
-        file_list = [u'.']
-    file_list = list(file_list)
-    return file_list
-
-
 class AddAction(object):
     """A class which defines what action to take when adding a file."""
 
@@ -150,6 +112,7 @@
 add_action_print = add_action_add_and_print
 
 
+@deprecated_function(zero_eighteen)
 def smart_add(file_list, recurse=True, action=None, save=True):
     """Add files to version, optionally recursing into directories.
 
@@ -157,256 +120,14 @@
     For the specific behaviour see the help for cmd_add().
 
     Returns the number of files added.
-    Please see smart_add_tree for more detail.
+    Deprecated in 0.18. Please use MutableTree.smart_add.
     """
-    file_list = _prepare_file_list(file_list)
     tree = WorkingTree.open_containing(file_list[0])[0]
     return smart_add_tree(tree, file_list, recurse, action=action, save=save)
 
 
-class FastPath(object):
-    """A path object with fast accessors for things like basename."""
-
-    __slots__ = ['raw_path', 'base_path']
-
-    def __init__(self, path, base_path=None):
-        """Construct a FastPath from path."""
-        if base_path is None:
-            self.base_path = bzrlib.osutils.basename(path)
-        else:
-            self.base_path = base_path
-        self.raw_path = path
-
-    def __cmp__(self, other):
-        return cmp(self.raw_path, other.raw_path)
-
-    def __hash__(self):
-        return hash(self.raw_path)
-
-
+@deprecated_function(zero_eighteen)
 def smart_add_tree(tree, file_list, recurse=True, action=None, save=True):
-    """Add files to version, optionally recursing into directories.
-
-    This is designed more towards DWIM for humans than API simplicity.
-    For the specific behaviour see the help for cmd_add().
-
-    This calls reporter with each (path, kind, file_id) of added files.
-
-    Returns the number of files added.
-
-    :param save: Save the inventory after completing the adds. If False this
-        provides dry-run functionality by doing the add and not saving the
-        inventory.  Note that the modified inventory is left in place, allowing
-        further dry-run tasks to take place. To restore the original inventory
-        call tree.read_working_inventory().
-    """
-    tree.lock_tree_write()
-    try:
-        return _smart_add_tree(tree=tree, file_list=file_list, recurse=recurse,
-                               action=action, save=save)
-    finally:
-        tree.unlock()
-
-
-def _smart_add_tree(tree, file_list, recurse=True, action=None, save=True):
-    """Helper for smart_add_tree.
-
-    The tree should be locked before entering this function. See smart_add_tree
-    for parameter definitions.
-    """
-    assert isinstance(recurse, bool)
-    if action is None:
-        action = AddAction()
-    
-    prepared_list = _prepare_file_list(file_list)
-    mutter("smart add of %r, originally %r", prepared_list, file_list)
-    inv = tree.read_working_inventory()
-    added = []
-    ignored = {}
-    dirs_to_add = []
-    user_dirs = set()
-
-    # validate user file paths and convert all paths to tree 
-    # relative : its cheaper to make a tree relative path an abspath
-    # than to convert an abspath to tree relative.
-    for filepath in prepared_list:
-        rf = FastPath(tree.relpath(filepath))
-        # validate user parameters. Our recursive code avoids adding new files
-        # that need such validation 
-        if tree.is_control_filename(rf.raw_path):
-            raise errors.ForbiddenControlFileError(filename=rf.raw_path)
-        
-        abspath = tree.abspath(rf.raw_path)
-        kind = bzrlib.osutils.file_kind(abspath)
-        if kind == 'directory':
-            # schedule the dir for scanning
-            user_dirs.add(rf)
-        else:
-            if not InventoryEntry.versionable_kind(kind):
-                raise errors.BadFileKindError(filename=abspath, kind=kind)
-        # ensure the named path is added, so that ignore rules in the later directory
-        # walk dont skip it.
-        # we dont have a parent ie known yet.: use the relatively slower inventory 
-        # probing method
-        versioned = inv.has_filename(rf.raw_path)
-        if versioned:
-            continue
-        added.extend(__add_one_and_parent(tree, inv, None, rf, kind, action))
-
-    if not recurse:
-        # no need to walk any directories at all.
-        if len(added) > 0 and save:
-            tree._write_inventory(inv)
-        return added, ignored
-
-    # only walk the minimal parents needed: we have user_dirs to override
-    # ignores.
-    prev_dir = None
-
-    is_inside = bzrlib.osutils.is_inside_or_parent_of_any
-    for path in sorted(user_dirs):
-        if (prev_dir is None or not is_inside([prev_dir], path.raw_path)):
-            dirs_to_add.append((path, None))
-        prev_dir = path.raw_path
-
-    # this will eventually be *just* directories, right now it starts off with 
-    # just directories.
-    for directory, parent_ie in dirs_to_add:
-        # directory is tree-relative
-        abspath = tree.abspath(directory.raw_path)
-
-        # get the contents of this directory.
-
-        # find the kind of the path being added.
-        kind = bzrlib.osutils.file_kind(abspath)
-
-        if not InventoryEntry.versionable_kind(kind):
-            warning("skipping %s (can't add file of kind '%s')", abspath, kind)
-            continue
-
-        if parent_ie is not None:
-            versioned = directory.base_path in parent_ie.children
-        else:
-            # without the parent ie, use the relatively slower inventory 
-            # probing method
-            versioned = inv.has_filename(directory.raw_path)
-
-        if kind == 'directory':
-            try:
-                sub_branch = bzrlib.bzrdir.BzrDir.open(abspath)
-                sub_tree = True
-            except NotBranchError:
-                sub_tree = False
-            except errors.UnsupportedFormatError:
-                sub_tree = True
-        else:
-            sub_tree = False
-
-        if directory.raw_path == '':
-            # mutter("tree root doesn't need to be added")
-            sub_tree = False
-        elif versioned:
-            pass
-            # mutter("%r is already versioned", abspath)
-        elif sub_tree:
-            # XXX: This is wrong; people *might* reasonably be trying to add
-            # subtrees as subtrees.  This should probably only be done in formats 
-            # which can represent subtrees, and even then perhaps only when
-            # the user asked to add subtrees.  At the moment you can add them
-            # specially through 'join --reference', which is perhaps
-            # reasonable: adding a new reference is a special operation and
-            # can have a special behaviour.  mbp 20070306
-            mutter("%r is a nested bzr tree", abspath)
-        else:
-            __add_one(tree, inv, parent_ie, directory, kind, action)
-            added.append(directory.raw_path)
-
-        if kind == 'directory' and not sub_tree:
-            if parent_ie is not None:
-                # must be present:
-                this_ie = parent_ie.children[directory.base_path]
-            else:
-                # without the parent ie, use the relatively slower inventory 
-                # probing method
-                this_id = inv.path2id(directory.raw_path)
-                if this_id is None:
-                    this_ie = None
-                else:
-                    this_ie = inv[this_id]
-
-            for subf in sorted(os.listdir(abspath)):
-                # here we could use TreeDirectory rather than 
-                # string concatenation.
-                subp = bzrlib.osutils.pathjoin(directory.raw_path, subf)
-                # TODO: is_control_filename is very slow. Make it faster. 
-                # TreeDirectory.is_control_filename could also make this 
-                # faster - its impossible for a non root dir to have a 
-                # control file.
-                if tree.is_control_filename(subp):
-                    mutter("skip control directory %r", subp)
-                elif subf in this_ie.children:
-                    # recurse into this already versioned subdir.
-                    dirs_to_add.append((FastPath(subp, subf), this_ie))
-                else:
-                    # user selection overrides ignoes
-                    # ignore while selecting files - if we globbed in the
-                    # outer loop we would ignore user files.
-                    ignore_glob = tree.is_ignored(subp)
-                    if ignore_glob is not None:
-                        # mutter("skip ignored sub-file %r", subp)
-                        ignored.setdefault(ignore_glob, []).append(subp)
-                    else:
-                        #mutter("queue to add sub-file %r", subp)
-                        dirs_to_add.append((FastPath(subp, subf), this_ie))
-
-    if len(added) > 0 and save:
-        tree._write_inventory(inv)
-    return added, ignored
-
-
-def __add_one_and_parent(tree, inv, parent_ie, path, kind, action):
-    """Add a new entry to the inventory and automatically add unversioned parents.
-
-    :param inv: Inventory which will receive the new entry.
-    :param parent_ie: Parent inventory entry if known, or None.  If
-        None, the parent is looked up by name and used if present, otherwise it
-        is recursively added.
-    :param kind: Kind of new entry (file, directory, etc)
-    :param action: callback(inv, parent_ie, path, kind); return ignored.
-    :return: A list of paths which have been added.
-    """
-    # Nothing to do if path is already versioned.
-    # This is safe from infinite recursion because the tree root is
-    # always versioned.
-    if parent_ie is not None:
-        # we have a parent ie already
-        added = []
-    else:
-        # slower but does not need parent_ie
-        if inv.has_filename(path.raw_path):
-            return []
-        # its really not there : add the parent
-        # note that the dirname use leads to some extra str copying etc but as
-        # there are a limited number of dirs we can be nested under, it should
-        # generally find it very fast and not recurse after that.
-        added = __add_one_and_parent(tree, inv, None, FastPath(dirname(path.raw_path)), 'directory', action)
-        parent_id = inv.path2id(dirname(path.raw_path))
-        parent_ie = inv[parent_id]
-    __add_one(tree, inv, parent_ie, path, kind, action)
-    return added + [path.raw_path]
-
-
-def __add_one(tree, inv, parent_ie, path, kind, action):
-    """Add a new entry to the inventory.
-
-    :param inv: Inventory which will receive the new entry.
-    :param parent_ie: Parent inventory entry.
-    :param kind: Kind of new entry (file, directory, etc)
-    :param action: callback(inv, parent_ie, path, kind); return a file_id 
-        or None to generate a new file id
-    :returns: None
-    """
-    file_id = action(inv, parent_ie, path, kind)
-    entry = bzrlib.inventory.make_entry(kind, path.base_path, parent_ie.file_id,
-                                        file_id=file_id)
-    inv.add(entry)
+    """Deprecated in 0.18. Please use MutableTree.smart_add."""
+    return tree.smart_add(file_list, recurse, action, save)
+

=== modified file 'bzrlib/benchmarks/tree_creator/kernel_like.py'
--- bzrlib/benchmarks/tree_creator/kernel_like.py	2007-04-09 22:11:22 +0000
+++ bzrlib/benchmarks/tree_creator/kernel_like.py	2007-07-04 02:43:26 +0000
@@ -20,7 +20,6 @@
 import os
 
 from bzrlib import (
-    add,
     bzrdir,
     osutils,
     workingtree,
@@ -119,7 +118,7 @@
         # Add everything to it
         tree.lock_write()
         try:
-            add.smart_add_tree(tree, [root], recurse=True, save=True)
+            tree.smart_add([root], recurse=True, save=True)
             if in_cache:
                 self._protect_files(root+'/.bzr')
         finally:

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2007-07-04 04:10:11 +0000
+++ bzrlib/builtins.py	2007-07-04 08:08:08 +0000
@@ -367,8 +367,13 @@
         if base_tree:
             base_tree.lock_read()
         try:
-            added, ignored = bzrlib.add.smart_add(file_list, not no_recurse,
-                action=action, save=not dry_run)
+            file_list = self._maybe_expand_globs(file_list)
+            if file_list:
+                tree = WorkingTree.open_containing(file_list[0])[0]
+            else:
+                tree = WorkingTree.open_containing(u'.')[0]
+            added, ignored = tree.smart_add(file_list, not
+                no_recurse, action=action, save=not dry_run)
         finally:
             if base_tree is not None:
                 base_tree.unlock()

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2007-06-25 14:27:08 +0000
+++ bzrlib/commands.py	2007-07-04 01:39:50 +0000
@@ -44,6 +44,7 @@
     option,
     osutils,
     trace,
+    win32utils,
     )
 """)
 
@@ -240,6 +241,19 @@
         if self.__doc__ == Command.__doc__:
             warn("No help message set for %r" % self)
 
+    def _maybe_expand_globs(self, file_list):
+        """Glob expand file_list if the platform does not do that itself.
+        
+        :return: A possibly empty list of unicode paths.
+
+        Introduced in bzrlib 0.18.
+        """
+        if not file_list:
+            file_list = []
+        if sys.platform == 'win32':
+            file_list = win32utils.glob_expand(file_list)
+        return list(file_list)
+
     def _usage(self):
         """Return single-line grammar for this command.
 

=== modified file 'bzrlib/generate_ids.py'
--- bzrlib/generate_ids.py	2007-02-18 00:22:24 +0000
+++ bzrlib/generate_ids.py	2007-07-02 22:36:11 +0000
@@ -47,7 +47,7 @@
     give a highly probably globally unique number. Then each call in the same
     process adds 1 to a serial number we append to that unique value.
     """
-    # XXX TODO: change bzrlib.add.smart_add to call workingtree.add() rather 
+    # XXX TODO: change bzrlib.add.smart_add_tree to call workingtree.add() rather 
     # than having to move the id randomness out of the inner loop like this.
     # XXX TODO: for the global randomness this uses we should add the thread-id
     # before the serial #.

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2007-06-21 03:29:39 +0000
+++ bzrlib/inventory.py	2007-07-03 02:33:32 +0000
@@ -1047,6 +1047,10 @@
                         child_dirs.append((child_relpath+'/', child_ie))
             stack.extend(reversed(child_dirs))
 
+    def make_entry(self, kind, name, parent_id, file_id=None):
+        """Simple thunk to bzrlib.inventory.make_entry."""
+        return make_entry(kind, name, parent_id, file_id)
+
     def entries(self):
         """Return list of (path, ie) for all entries except the root.
 

=== modified file 'bzrlib/mutabletree.py'
--- bzrlib/mutabletree.py	2007-04-17 20:10:04 +0000
+++ bzrlib/mutabletree.py	2007-07-04 05:50:05 +0000
@@ -20,6 +20,18 @@
 """
 
 
+from bzrlib.lazy_import import lazy_import
+lazy_import(globals(), """
+import os
+
+from bzrlib import (
+    add,
+    bzrdir,
+    )
+from bzrlib.osutils import dirname
+from bzrlib.trace import mutter, warning
+""")
+
 from bzrlib import (
     errors,
     osutils,
@@ -260,3 +272,247 @@
             parent tree - i.e. a ghost.
         """
         raise NotImplementedError(self.set_parent_trees)
+
+    @needs_tree_write_lock
+    def smart_add(self, file_list, recurse=True, action=None, save=True):
+        """Version file_list, optionally recursing into directories.
+
+        This is designed more towards DWIM for humans than API clarity.
+        For the specific behaviour see the help for cmd_add().
+
+        :param action: A reporter to be called with the inventory, parent_ie,
+            path and kind of the path being added. It may return a file_id if 
+            a specific one should be used.
+        :param save: Save the inventory after completing the adds. If False
+            this provides dry-run functionality by doing the add and not saving
+            the inventory.  Note that the modified inventory is left in place,
+            allowing further dry-run tasks to take place. To restore the
+            original inventory call self.read_working_inventory().
+        :return: A tuple - files_added, ignored_files. files_added is the count
+            of added files, and ignored_files is a dict mapping files that were
+            ignored to the rule that caused them to be ignored.
+        """
+        # not in an inner loop; and we want to remove direct use of this,
+        # so here as a reminder for now. RBC 20070703
+        from bzrlib.inventory import InventoryEntry
+        assert isinstance(recurse, bool)
+        if action is None:
+            action = add.AddAction()
+        
+        if not file_list:
+            # no paths supplied: add the entire tree.
+            file_list = [u'.']
+        # mutter("smart add of %r")
+        inv = self.inventory
+        added = []
+        ignored = {}
+        dirs_to_add = []
+        user_dirs = set()
+
+        # validate user file paths and convert all paths to tree 
+        # relative : it's cheaper to make a tree relative path an abspath
+        # than to convert an abspath to tree relative.
+        for filepath in file_list:
+            rf = _FastPath(self.relpath(filepath))
+            # validate user parameters. Our recursive code avoids adding new files
+            # that need such validation 
+            if self.is_control_filename(rf.raw_path):
+                raise errors.ForbiddenControlFileError(filename=rf.raw_path)
+            
+            abspath = self.abspath(rf.raw_path)
+            kind = osutils.file_kind(abspath)
+            if kind == 'directory':
+                # schedule the dir for scanning
+                user_dirs.add(rf)
+            else:
+                if not InventoryEntry.versionable_kind(kind):
+                    raise errors.BadFileKindError(filename=abspath, kind=kind)
+            # ensure the named path is added, so that ignore rules in the later directory
+            # walk dont skip it.
+            # we dont have a parent ie known yet.: use the relatively slower inventory 
+            # probing method
+            versioned = inv.has_filename(rf.raw_path)
+            if versioned:
+                continue
+            added.extend(_add_one_and_parent(self, inv, None, rf, kind, action))
+
+        if not recurse:
+            # no need to walk any directories at all.
+            if len(added) > 0 and save:
+                self._write_inventory(inv)
+            return added, ignored
+
+        # only walk the minimal parents needed: we have user_dirs to override
+        # ignores.
+        prev_dir = None
+
+        is_inside = osutils.is_inside_or_parent_of_any
+        for path in sorted(user_dirs):
+            if (prev_dir is None or not is_inside([prev_dir], path.raw_path)):
+                dirs_to_add.append((path, None))
+            prev_dir = path.raw_path
+
+        # dirs_to_add is initialised to a list of directories, but as we scan
+        # directories we append files to it.
+        # XXX: We should determine kind of files when we scan them rather than
+        # adding to this list. RBC 20070703
+        for directory, parent_ie in dirs_to_add:
+            # directory is tree-relative
+            abspath = self.abspath(directory.raw_path)
+
+            # get the contents of this directory.
+
+            # find the kind of the path being added.
+            kind = osutils.file_kind(abspath)
+
+            if not InventoryEntry.versionable_kind(kind):
+                warning("skipping %s (can't add file of kind '%s')", abspath, kind)
+                continue
+
+            if parent_ie is not None:
+                versioned = directory.base_path in parent_ie.children
+            else:
+                # without the parent ie, use the relatively slower inventory 
+                # probing method
+                versioned = inv.has_filename(directory.raw_path)
+
+            if kind == 'directory':
+                try:
+                    sub_branch = bzrdir.BzrDir.open(abspath)
+                    sub_tree = True
+                except errors.NotBranchError:
+                    sub_tree = False
+                except errors.UnsupportedFormatError:
+                    sub_tree = True
+            else:
+                sub_tree = False
+
+            if directory.raw_path == '':
+                # mutter("tree root doesn't need to be added")
+                sub_tree = False
+            elif versioned:
+                pass
+                # mutter("%r is already versioned", abspath)
+            elif sub_tree:
+                # XXX: This is wrong; people *might* reasonably be trying to add
+                # subtrees as subtrees.  This should probably only be done in formats 
+                # which can represent subtrees, and even then perhaps only when
+                # the user asked to add subtrees.  At the moment you can add them
+                # specially through 'join --reference', which is perhaps
+                # reasonable: adding a new reference is a special operation and
+                # can have a special behaviour.  mbp 20070306
+                mutter("%r is a nested bzr tree", abspath)
+            else:
+                _add_one(self, inv, parent_ie, directory, kind, action)
+                added.append(directory.raw_path)
+
+            if kind == 'directory' and not sub_tree:
+                if parent_ie is not None:
+                    # must be present:
+                    this_ie = parent_ie.children[directory.base_path]
+                else:
+                    # without the parent ie, use the relatively slower inventory 
+                    # probing method
+                    this_id = inv.path2id(directory.raw_path)
+                    if this_id is None:
+                        this_ie = None
+                    else:
+                        this_ie = inv[this_id]
+
+                for subf in sorted(os.listdir(abspath)):
+                    # here we could use TreeDirectory rather than 
+                    # string concatenation.
+                    subp = osutils.pathjoin(directory.raw_path, subf)
+                    # TODO: is_control_filename is very slow. Make it faster. 
+                    # TreeDirectory.is_control_filename could also make this 
+                    # faster - its impossible for a non root dir to have a 
+                    # control file.
+                    if self.is_control_filename(subp):
+                        mutter("skip control directory %r", subp)
+                    elif subf in this_ie.children:
+                        # recurse into this already versioned subdir.
+                        dirs_to_add.append((_FastPath(subp, subf), this_ie))
+                    else:
+                        # user selection overrides ignoes
+                        # ignore while selecting files - if we globbed in the
+                        # outer loop we would ignore user files.
+                        ignore_glob = self.is_ignored(subp)
+                        if ignore_glob is not None:
+                            # mutter("skip ignored sub-file %r", subp)
+                            ignored.setdefault(ignore_glob, []).append(subp)
+                        else:
+                            #mutter("queue to add sub-file %r", subp)
+                            dirs_to_add.append((_FastPath(subp, subf), this_ie))
+
+        if len(added) > 0 and save:
+            self._write_inventory(inv)
+        return added, ignored
+
+
+class _FastPath(object):
+    """A path object with fast accessors for things like basename."""
+
+    __slots__ = ['raw_path', 'base_path']
+
+    def __init__(self, path, base_path=None):
+        """Construct a FastPath from path."""
+        if base_path is None:
+            self.base_path = osutils.basename(path)
+        else:
+            self.base_path = base_path
+        self.raw_path = path
+
+    def __cmp__(self, other):
+        return cmp(self.raw_path, other.raw_path)
+
+    def __hash__(self):
+        return hash(self.raw_path)
+
+
+def _add_one_and_parent(tree, inv, parent_ie, path, kind, action):
+    """Add a new entry to the inventory and automatically add unversioned parents.
+
+    :param inv: Inventory which will receive the new entry.
+    :param parent_ie: Parent inventory entry if known, or None.  If
+        None, the parent is looked up by name and used if present, otherwise it
+        is recursively added.
+    :param kind: Kind of new entry (file, directory, etc)
+    :param action: callback(inv, parent_ie, path, kind); return ignored.
+    :return: A list of paths which have been added.
+    """
+    # Nothing to do if path is already versioned.
+    # This is safe from infinite recursion because the tree root is
+    # always versioned.
+    if parent_ie is not None:
+        # we have a parent ie already
+        added = []
+    else:
+        # slower but does not need parent_ie
+        if inv.has_filename(path.raw_path):
+            return []
+        # its really not there : add the parent
+        # note that the dirname use leads to some extra str copying etc but as
+        # there are a limited number of dirs we can be nested under, it should
+        # generally find it very fast and not recurse after that.
+        added = _add_one_and_parent(tree, inv, None,
+            _FastPath(dirname(path.raw_path)), 'directory', action)
+        parent_id = inv.path2id(dirname(path.raw_path))
+        parent_ie = inv[parent_id]
+    _add_one(tree, inv, parent_ie, path, kind, action)
+    return added + [path.raw_path]
+
+
+def _add_one(tree, inv, parent_ie, path, kind, file_id_callback):
+    """Add a new entry to the inventory.
+
+    :param inv: Inventory which will receive the new entry.
+    :param parent_ie: Parent inventory entry.
+    :param kind: Kind of new entry (file, directory, etc)
+    :param file_id_callback: callback(inv, parent_ie, path, kind); return a
+        file_id or None to generate a new file id
+    :returns: None
+    """
+    file_id = file_id_callback(inv, parent_ie, path, kind)
+    entry = inv.make_entry(kind, path.base_path, parent_ie.file_id,
+        file_id=file_id)
+    inv.add(entry)

=== modified file 'bzrlib/tests/blackbox/test_add.py'
--- bzrlib/tests/blackbox/test_add.py	2007-06-26 10:12:33 +0000
+++ bzrlib/tests/blackbox/test_add.py	2007-07-03 02:33:32 +0000
@@ -175,13 +175,12 @@
         self.assertEqual(base_tree.path2id('b/d'), new_tree.path2id('d'))
 
     def test_add_dry_run(self):
-        # ensure that --dry-run actually don't add anything
-        base_tree = self.make_branch_and_tree('.')
-        self.build_tree(['spam'])
-        out = self.run_bzr(['add', '--dry-run'], retcode=0)[0]
-        self.assertEquals('added spam\n', out)
-        out = self.run_bzr(['added'], retcode=0)[0]
-        self.assertEquals('', out)
+        """Test a dry run add, make sure nothing is added."""
+        wt = self.make_branch_and_tree('.')
+        self.build_tree(['inertiatic/', 'inertiatic/esp'])
+        self.assertEqual(list(wt.unknowns()), ['inertiatic'])
+        self.run_bzr('add --dry-run')
+        self.assertEqual(list(wt.unknowns()), ['inertiatic'])
 
     def test_add_control_dir(self):
         """The control dir and its content should be refused."""

=== modified file 'bzrlib/tests/test_bad_files.py'
--- bzrlib/tests/test_bad_files.py	2007-02-28 22:40:54 +0000
+++ bzrlib/tests/test_bad_files.py	2007-07-04 02:43:26 +0000
@@ -70,8 +70,7 @@
                            ])
 
         # We should raise an error if we are adding a bogus file
-        self.assertRaises(errors.BadFileKindError,
-                          add.smart_add_tree, wt, ['a-fifo'])
+        self.assertRaises(errors.BadFileKindError, wt.smart_add, ['a-fifo'])
 
         # And the list of files shouldn't have been modified
         verify_status(self, wt,
@@ -82,7 +81,7 @@
 
         # Make sure smart_add can handle having a bogus
         # file in the way
-        add.smart_add_tree(wt, ['.'])
+        wt.smart_add([])
         verify_status(self, wt,
                           ['added:\n',
                            '  six\n',

=== modified file 'bzrlib/tests/test_merge_core.py'
--- bzrlib/tests/test_merge_core.py	2007-06-26 08:46:23 +0000
+++ bzrlib/tests/test_merge_core.py	2007-07-04 02:43:26 +0000
@@ -23,7 +23,6 @@
     generate_ids,
     osutils,
     )
-from bzrlib.add import smart_add_tree
 from bzrlib.builtins import merge
 from bzrlib.conflicts import ContentsConflict, TextConflict, PathConflict
 from bzrlib.errors import (NotBranchError, NotVersionedError,
@@ -426,7 +425,7 @@
         self.build_tree(("original/", "original/file1", "original/file2"))
         tree = self.make_branch_and_tree('original')
         branch = tree.branch
-        smart_add_tree(tree, ["original"])
+        tree.smart_add(["original"])
         tree.commit("start branch.", verbose=False)
         # Mary branches it.
         self.build_tree(("mary/",))

=== modified file 'bzrlib/tests/test_smart_add.py'
--- bzrlib/tests/test_smart_add.py	2007-06-26 10:12:33 +0000
+++ bzrlib/tests/test_smart_add.py	2007-07-03 02:38:36 +0000
@@ -1,4 +1,4 @@
-# Copyright (C) 2005, 2006 Canonical Ltd
+# Copyright (C) 2005, 2006, 2007 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -15,128 +15,14 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 from cStringIO import StringIO
-import os
-import unittest
 
-from bzrlib import errors, ignores, osutils
+from bzrlib import osutils
 from bzrlib.add import (
     AddAction,
     AddFromBaseAction,
-    smart_add,
-    smart_add_tree,
     )
-from bzrlib.tests import TestCase, TestCaseWithTransport, TestSkipped
-from bzrlib.errors import NoSuchFile
-from bzrlib.inventory import InventoryFile, Inventory
-from bzrlib.workingtree import WorkingTree
-
-
-class TestSmartAdd(TestCaseWithTransport):
-
-    def test_add_dot_from_root(self):
-        """Test adding . from the root of the tree.""" 
-        paths = ("original/", "original/file1", "original/file2")
-        self.build_tree(paths)
-        wt = self.make_branch_and_tree('.')
-        smart_add_tree(wt, (u".",))
-        for path in paths:
-            self.assertNotEqual(wt.path2id(path), None)
-
-    def test_add_dot_from_subdir(self):
-        """Test adding . from a subdir of the tree.""" 
-        from bzrlib.add import smart_add
-        paths = ("original/", "original/file1", "original/file2")
-        self.build_tree(paths)
-        wt = self.make_branch_and_tree('.')
-        os.chdir("original")
-        smart_add_tree(wt, (u".",))
-        for path in paths:
-            self.assertNotEqual(wt.path2id(path), None)
-
-    def test_add_tree_from_above_tree(self):
-        """Test adding a tree from above the tree.""" 
-        from bzrlib.add import smart_add
-        paths = ("original/", "original/file1", "original/file2")
-        branch_paths = ("branch/", "branch/original/", "branch/original/file1",
-                        "branch/original/file2")
-        self.build_tree(branch_paths)
-        wt = self.make_branch_and_tree('branch')
-        smart_add_tree(wt, ("branch",))
-        for path in paths:
-            self.assertNotEqual(wt.path2id(path), None)
-
-    def test_add_above_tree_preserves_tree(self):
-        """Test nested trees are not affect by an add above them."""
-        from bzrlib.add import smart_add
-        paths = ("original/", "original/file1", "original/file2")
-        child_paths = ("path",)
-        full_child_paths = ("original/child", "original/child/path")
-        build_paths = ("original/", "original/file1", "original/file2", 
-                       "original/child/", "original/child/path")
-        
-        self.build_tree(build_paths)
-        wt = self.make_branch_and_tree('.')
-        child_tree = self.make_branch_and_tree('original/child')
-        smart_add_tree(wt, (".",))
-        for path in paths:
-            self.assertNotEqual((path, wt.path2id(path)),
-                                (path, None))
-        for path in full_child_paths:
-            self.assertEqual((path, wt.path2id(path)),
-                             (path, None))
-        for path in child_paths:
-            self.assertEqual(child_tree.path2id(path), None)
-
-    def test_add_paths(self):
-        """Test smart-adding a list of paths."""
-        from bzrlib.add import smart_add
-        paths = ("file1", "file2")
-        self.build_tree(paths)
-        wt = self.make_branch_and_tree('.')
-        smart_add_tree(wt, paths)
-        for path in paths:
-            self.assertNotEqual(wt.path2id(path), None)
-    
-    def test_add_ignored_nested_paths(self):
-        """Test smart-adding a list of paths which includes ignored ones."""
-        wt = self.make_branch_and_tree('.')
-        tree_shape = ("adir/", "adir/CVS/", "adir/CVS/afile", "adir/CVS/afile2")
-        add_paths = ("adir/CVS", "adir/CVS/afile", "adir")
-        expected_paths = ("adir", "adir/CVS", "adir/CVS/afile", "adir/CVS/afile2")
-        self.build_tree(tree_shape)
-        smart_add_tree(wt, add_paths)
-        for path in expected_paths:
-            self.assertNotEqual(wt.path2id(path), None, "No id added for %s" % path)
-
-    def test_add_dry_run(self):
-        """Test a dry run add, make sure nothing is added."""
-        from bzrlib.commands import run_bzr
-        eq = self.assertEqual
-        wt = self.make_branch_and_tree('.')
-        self.build_tree(['inertiatic/', 'inertiatic/esp'])
-        eq(list(wt.unknowns()), ['inertiatic'])
-        self.run_bzr('add --dry-run .')
-        eq(list(wt.unknowns()), ['inertiatic'])
-
-    def test_add_non_existant(self):
-        """Test smart-adding a file that does not exist."""
-        from bzrlib.add import smart_add
-        wt = self.make_branch_and_tree('.')
-        self.assertRaises(NoSuchFile, smart_add_tree, wt, 'non-existant-file')
-
-    def test_returns_and_ignores(self):
-        """Correctly returns added/ignored files"""
-        from bzrlib.commands import run_bzr
-        wt = self.make_branch_and_tree('.')
-        # The default ignore list includes '*.py[co]', but not CVS
-        ignores._set_user_ignores(['*.py[co]'])
-        self.build_tree(['inertiatic/', 'inertiatic/esp', 'inertiatic/CVS',
-                        'inertiatic/foo.pyc'])
-        added, ignored = smart_add_tree(wt, u'.')
-        self.assertSubset(('inertiatic', 'inertiatic/esp', 'inertiatic/CVS'),
-                          added)
-        self.assertSubset(('*.py[co]',), ignored)
-        self.assertSubset(('inertiatic/foo.pyc',), ignored['*.py[co]'])
+from bzrlib.tests import TestCase, TestCaseWithTransport
+from bzrlib.inventory import Inventory
 
 
 class AddCustomIDAction(AddAction):
@@ -153,113 +39,6 @@
         return file_id
 
 
-class TestSmartAddTree(TestCaseWithTransport):
-    """Test smart adds with a specified branch."""
-
-    def test_add_dot_from_root(self):
-        """Test adding . from the root of the tree.""" 
-        paths = ("original/", "original/file1", "original/file2")
-        self.build_tree(paths)
-        wt = self.make_branch_and_tree('.')
-        smart_add_tree(wt, (u".",))
-        for path in paths:
-            self.assertNotEqual(wt.path2id(path), None)
-
-    def test_add_dot_from_subdir(self):
-        """Test adding . from a subdir of the tree.""" 
-        paths = ("original/", "original/file1", "original/file2")
-        self.build_tree(paths)
-        wt = self.make_branch_and_tree('.')
-        os.chdir("original")
-        smart_add_tree(wt, (u".",))
-        for path in paths:
-            self.assertNotEqual(wt.path2id(path), None)
-
-    def test_add_tree_from_above_tree(self):
-        """Test adding a tree from above the tree.""" 
-        paths = ("original/", "original/file1", "original/file2")
-        branch_paths = ("branch/", "branch/original/", "branch/original/file1",
-                        "branch/original/file2")
-        self.build_tree(branch_paths)
-        tree = self.make_branch_and_tree('branch')
-        smart_add_tree(tree, ("branch",))
-        for path in paths:
-            self.assertNotEqual(tree.path2id(path), None)
-
-    def test_add_above_tree_preserves_tree(self):
-        """Test nested trees are not affect by an add above them."""
-        paths = ("original/", "original/file1", "original/file2")
-        child_paths = ("path")
-        full_child_paths = ("original/child", "original/child/path")
-        build_paths = ("original/", "original/file1", "original/file2", 
-                       "original/child/", "original/child/path")
-        self.build_tree(build_paths)
-        tree = self.make_branch_and_tree('.')
-        child_tree = self.make_branch_and_tree("original/child")
-        smart_add_tree(tree, (u".",))
-        for path in paths:
-            self.assertNotEqual((path, tree.path2id(path)),
-                                (path, None))
-        for path in full_child_paths:
-            self.assertEqual((path, tree.path2id(path)),
-                             (path, None))
-        for path in child_paths:
-            self.assertEqual(child_tree.path2id(path), None)
-
-    def test_add_paths(self):
-        """Test smart-adding a list of paths."""
-        paths = ("file1", "file2")
-        self.build_tree(paths)
-        wt = self.make_branch_and_tree('.')
-        smart_add_tree(wt, paths)
-        for path in paths:
-            self.assertNotEqual(wt.path2id(path), None)
-
-    def test_add_multiple_dirs(self):
-        """Test smart adding multiple directories at once."""
-        added_paths = ['file1', 'file2',
-                       'dir1/', 'dir1/file3',
-                       'dir1/subdir2/', 'dir1/subdir2/file4',
-                       'dir2/', 'dir2/file5',
-                      ]
-        not_added = ['file6', 'dir3/', 'dir3/file7', 'dir3/file8']
-        self.build_tree(added_paths)
-        self.build_tree(not_added)
-
-        wt = self.make_branch_and_tree('.')
-        smart_add_tree(wt, ['file1', 'file2', 'dir1', 'dir2'])
-
-        for path in added_paths:
-            self.assertNotEqual(None, wt.path2id(path.rstrip('/')),
-                    'Failed to add path: %s' % (path,))
-        for path in not_added:
-            self.assertEqual(None, wt.path2id(path.rstrip('/')),
-                    'Accidentally added path: %s' % (path,))
-
-    def test_custom_ids(self):
-        sio = StringIO()
-        action = AddCustomIDAction(to_file=sio, should_print=True)
-        self.build_tree(['file1', 'dir1/', 'dir1/file2'])
-
-        wt = self.make_branch_and_tree('.')
-        smart_add_tree(wt, ['.'], action=action)
-        # The order of adds is not strictly fixed:
-        sio.seek(0)
-        lines = sorted(sio.readlines())
-        self.assertEqualDiff(['added dir1 with id directory-dir1\n',
-                              'added dir1/file2 with id file-dir1%file2\n',
-                              'added file1 with id file-file1\n',
-                             ], lines)
-        wt.lock_read()
-        self.addCleanup(wt.unlock)
-        self.assertEqual([('', wt.path2id('')),
-                          ('dir1', 'directory-dir1'),
-                          ('dir1/file2', 'file-dir1%file2'),
-                          ('file1', 'file-file1'),
-                         ], [(path, ie.file_id) for path, ie
-                                in wt.inventory.iter_entries()])
-
-
 class TestAddFrom(TestCaseWithTransport):
     """Tests for AddFromBaseAction"""
 
@@ -284,7 +63,7 @@
                 action = AddFromBaseAction(base_tree, base_path,
                                            to_file=to_file,
                                            should_print=should_print)
-                smart_add_tree(new_tree, file_list, action=action)
+                new_tree.smart_add(file_list, action=action)
             finally:
                 new_tree.unlock()
         finally:
@@ -365,68 +144,6 @@
         self.failIf(a_id in self.base_tree)
 
 
-class TestAddNonNormalized(TestCaseWithTransport):
-
-    def make(self):
-        try:
-            self.build_tree([u'a\u030a'])
-        except UnicodeError:
-            raise TestSkipped('Filesystem cannot create unicode filenames')
-
-        self.wt = self.make_branch_and_tree('.')
-
-    def test_accessible_explicit(self):
-        self.make()
-        orig = osutils.normalized_filename
-        osutils.normalized_filename = osutils._accessible_normalized_filename
-        try:
-            smart_add_tree(self.wt, [u'a\u030a'])
-            self.wt.lock_read()
-            self.addCleanup(self.wt.unlock)
-            self.assertEqual([('', 'directory'), (u'\xe5', 'file')],
-                    [(path, ie.kind) for path,ie in 
-                        self.wt.inventory.iter_entries()])
-        finally:
-            osutils.normalized_filename = orig
-
-    def test_accessible_implicit(self):
-        self.make()
-        orig = osutils.normalized_filename
-        osutils.normalized_filename = osutils._accessible_normalized_filename
-        try:
-            smart_add_tree(self.wt, [])
-            self.wt.lock_read()
-            self.addCleanup(self.wt.unlock)
-            self.assertEqual([('', 'directory'), (u'\xe5', 'file')],
-                    [(path, ie.kind) for path,ie in 
-                        self.wt.inventory.iter_entries()])
-        finally:
-            osutils.normalized_filename = orig
-
-    def test_inaccessible_explicit(self):
-        self.make()
-        orig = osutils.normalized_filename
-        osutils.normalized_filename = osutils._inaccessible_normalized_filename
-        try:
-            self.assertRaises(errors.InvalidNormalization,
-                    smart_add_tree, self.wt, [u'a\u030a'])
-        finally:
-            osutils.normalized_filename = orig
-
-    def test_inaccessible_implicit(self):
-        self.make()
-        orig = osutils.normalized_filename
-        osutils.normalized_filename = osutils._inaccessible_normalized_filename
-        try:
-            # TODO: jam 20060701 In the future, this should probably
-            #       just ignore files that don't fit the normalization
-            #       rules, rather than exploding
-            self.assertRaises(errors.InvalidNormalization,
-                    smart_add_tree, self.wt, [])
-        finally:
-            osutils.normalized_filename = orig
-
-
 class TestAddActions(TestCase):
 
     def test_quiet(self):
@@ -436,10 +153,12 @@
         self.run_action("added path\n")
 
     def run_action(self, output):
-        from bzrlib.add import AddAction, FastPath
+        from bzrlib.add import AddAction
+        from bzrlib.mutabletree import _FastPath
         inv = Inventory()
         stdout = StringIO()
         action = AddAction(to_file=stdout, should_print=bool(output))
 
-        self.apply_redirected(None, stdout, None, action, inv, None, FastPath('path'), 'file')
+        self.apply_redirected(None, stdout, None, action, inv, None,
+            _FastPath('path'), 'file')
         self.assertEqual(stdout.getvalue(), output)

=== modified file 'bzrlib/tests/workingtree_implementations/test_smart_add.py'
--- bzrlib/tests/workingtree_implementations/test_smart_add.py	2007-03-02 07:57:54 +0000
+++ bzrlib/tests/workingtree_implementations/test_smart_add.py	2007-07-04 06:25:03 +0000
@@ -16,18 +16,21 @@
 
 """Test that we can use smart_add on all Tree implementations."""
 
+from cStringIO import StringIO
+
 from bzrlib import (
     add,
     errors,
+    ignores,
+    osutils,
+    tests,
     workingtree,
     )
 from bzrlib.add import (
     AddAction,
     AddFromBaseAction,
-    smart_add,
-    smart_add_tree,
     )
-
+from bzrlib.tests.test_smart_add import AddCustomIDAction
 from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
 
 
@@ -36,7 +39,7 @@
     def test_single_file(self):
         tree = self.make_branch_and_tree('tree')
         self.build_tree(['tree/a'])
-        add.smart_add_tree(tree, ['tree'])
+        tree.smart_add(['tree'])
 
         tree.lock_read()
         try:
@@ -52,9 +55,199 @@
         """Dry-run add doesn't permanently affect the tree."""
         wt = self.make_branch_and_tree('.')
         self.build_tree(['file'])
-        smart_add_tree(wt, ['file'], save=False)
-        # The in-memory inventory is left modified in inventory-based trees;
-        # it may not be in dirstate trees.  Anyhow, now we reload to make sure
-        # the on-disk version is not modified.
+        wt.smart_add(['file'], save=False)
+        # the file should not be added - no id.
+        self.assertEqual(wt.path2id('file'), None)
+        # and the disk state should be the same - reopen to check.
         wt = wt.bzrdir.open_workingtree()
         self.assertEqual(wt.path2id('file'), None)
+
+    def test_add_dot_from_root(self):
+        """Test adding . from the root of the tree.""" 
+        paths = ("original/", "original/file1", "original/file2")
+        self.build_tree(paths)
+        wt = self.make_branch_and_tree('.')
+        wt.smart_add((u".",))
+        for path in paths:
+            self.assertNotEqual(wt.path2id(path), None)
+
+    def test_add_dot_from_subdir(self):
+        """Test adding . from a subdir of the tree.""" 
+        paths = ("original/", "original/file1", "original/file2")
+        self.build_tree(paths)
+        wt = self.make_branch_and_tree('.')
+        wt.smart_add((u".",))
+        for path in paths:
+            self.assertNotEqual(wt.path2id(path), None)
+
+    def test_add_tree_from_above_tree(self):
+        """Test adding a tree from above the tree.""" 
+        paths = ("original/", "original/file1", "original/file2")
+        branch_paths = ("branch/", "branch/original/", "branch/original/file1",
+                        "branch/original/file2")
+        self.build_tree(branch_paths)
+        wt = self.make_branch_and_tree('branch')
+        wt.smart_add(("branch",))
+        for path in paths:
+            self.assertNotEqual(wt.path2id(path), None)
+
+    def test_add_above_tree_preserves_tree(self):
+        """Test nested trees are not affect by an add above them."""
+        paths = ("original/", "original/file1", "original/file2")
+        child_paths = ("path",)
+        full_child_paths = ("original/child", "original/child/path")
+        build_paths = ("original/", "original/file1", "original/file2", 
+                       "original/child/", "original/child/path")
+        
+        self.build_tree(build_paths)
+        wt = self.make_branch_and_tree('.')
+        child_tree = self.make_branch_and_tree('original/child')
+        wt.smart_add((".",))
+        for path in paths:
+            self.assertNotEqual((path, wt.path2id(path)),
+                                (path, None))
+        for path in full_child_paths:
+            self.assertEqual((path, wt.path2id(path)),
+                             (path, None))
+        for path in child_paths:
+            self.assertEqual(child_tree.path2id(path), None)
+
+    def test_add_paths(self):
+        """Test smart-adding a list of paths."""
+        paths = ("file1", "file2")
+        self.build_tree(paths)
+        wt = self.make_branch_and_tree('.')
+        wt.smart_add(paths)
+        for path in paths:
+            self.assertNotEqual(wt.path2id(path), None)
+    
+    def test_add_ignored_nested_paths(self):
+        """Test smart-adding a list of paths which includes ignored ones."""
+        wt = self.make_branch_and_tree('.')
+        tree_shape = ("adir/", "adir/CVS/", "adir/CVS/afile", "adir/CVS/afile2")
+        add_paths = ("adir/CVS", "adir/CVS/afile", "adir")
+        expected_paths = ("adir", "adir/CVS", "adir/CVS/afile", "adir/CVS/afile2")
+        self.build_tree(tree_shape)
+        wt.smart_add(add_paths)
+        for path in expected_paths:
+            self.assertNotEqual(wt.path2id(path), None, "No id added for %s" % path)
+
+    def test_add_non_existant(self):
+        """Test smart-adding a file that does not exist."""
+        wt = self.make_branch_and_tree('.')
+        self.assertRaises(errors.NoSuchFile, wt.smart_add, ['non-existant-file'])
+
+    def test_returns_and_ignores(self):
+        """Correctly returns added/ignored files"""
+        wt = self.make_branch_and_tree('.')
+        # The default ignore list includes '*.py[co]', but not CVS
+        ignores._set_user_ignores(['*.py[co]'])
+        self.build_tree(['inertiatic/', 'inertiatic/esp', 'inertiatic/CVS',
+                        'inertiatic/foo.pyc'])
+        added, ignored = wt.smart_add(u'.')
+        self.assertSubset(('inertiatic', 'inertiatic/esp', 'inertiatic/CVS'),
+                          added)
+        self.assertSubset(('*.py[co]',), ignored)
+        self.assertSubset(('inertiatic/foo.pyc',), ignored['*.py[co]'])
+
+    def test_add_multiple_dirs(self):
+        """Test smart adding multiple directories at once."""
+        added_paths = ['file1', 'file2',
+                       'dir1/', 'dir1/file3',
+                       'dir1/subdir2/', 'dir1/subdir2/file4',
+                       'dir2/', 'dir2/file5',
+                      ]
+        not_added = ['file6', 'dir3/', 'dir3/file7', 'dir3/file8']
+        self.build_tree(added_paths)
+        self.build_tree(not_added)
+
+        wt = self.make_branch_and_tree('.')
+        wt.smart_add(['file1', 'file2', 'dir1', 'dir2'])
+
+        for path in added_paths:
+            self.assertNotEqual(None, wt.path2id(path.rstrip('/')),
+                    'Failed to add path: %s' % (path,))
+        for path in not_added:
+            self.assertEqual(None, wt.path2id(path.rstrip('/')),
+                    'Accidentally added path: %s' % (path,))
+
+    def test_custom_ids(self):
+        sio = StringIO()
+        action = AddCustomIDAction(to_file=sio, should_print=True)
+        self.build_tree(['file1', 'dir1/', 'dir1/file2'])
+
+        wt = self.make_branch_and_tree('.')
+        wt.smart_add(['.'], action=action)
+        # The order of adds is not strictly fixed:
+        sio.seek(0)
+        lines = sorted(sio.readlines())
+        self.assertEqualDiff(['added dir1 with id directory-dir1\n',
+                              'added dir1/file2 with id file-dir1%file2\n',
+                              'added file1 with id file-file1\n',
+                             ], lines)
+        wt.lock_read()
+        self.addCleanup(wt.unlock)
+        self.assertEqual([('', wt.path2id('')),
+                          ('dir1', 'directory-dir1'),
+                          ('dir1/file2', 'file-dir1%file2'),
+                          ('file1', 'file-file1'),
+                         ], [(path, ie.file_id) for path, ie
+                                in wt.inventory.iter_entries()])
+
+    def make_unicode_containing_tree(self):
+        try:
+            self.build_tree([u'a\u030a'])
+        except UnicodeError:
+            raise tests.TestSkipped('Filesystem cannot create unicode filenames')
+        self.wt = self.make_branch_and_tree('.')
+
+    def test_accessible_explicit(self):
+        self.make_unicode_containing_tree()
+        orig = osutils.normalized_filename
+        osutils.normalized_filename = osutils._accessible_normalized_filename
+        try:
+            self.wt.smart_add([u'a\u030a'])
+            self.wt.lock_read()
+            self.addCleanup(self.wt.unlock)
+            self.assertEqual([('', 'directory'), (u'\xe5', 'file')],
+                    [(path, ie.kind) for path,ie in 
+                        self.wt.inventory.iter_entries()])
+        finally:
+            osutils.normalized_filename = orig
+
+    def test_accessible_implicit(self):
+        self.make_unicode_containing_tree()
+        orig = osutils.normalized_filename
+        osutils.normalized_filename = osutils._accessible_normalized_filename
+        try:
+            self.wt.smart_add([])
+            self.wt.lock_read()
+            self.addCleanup(self.wt.unlock)
+            self.assertEqual([('', 'directory'), (u'\xe5', 'file')],
+                    [(path, ie.kind) for path,ie in 
+                        self.wt.inventory.iter_entries()])
+        finally:
+            osutils.normalized_filename = orig
+
+    def test_inaccessible_explicit(self):
+        self.make_unicode_containing_tree()
+        orig = osutils.normalized_filename
+        osutils.normalized_filename = osutils._inaccessible_normalized_filename
+        try:
+            self.assertRaises(errors.InvalidNormalization,
+                    self.wt.smart_add, [u'a\u030a'])
+        finally:
+            osutils.normalized_filename = orig
+
+    def test_inaccessible_implicit(self):
+        self.make_unicode_containing_tree()
+        orig = osutils.normalized_filename
+        osutils.normalized_filename = osutils._inaccessible_normalized_filename
+        try:
+            # TODO: jam 20060701 In the future, this should probably
+            #       just ignore files that don't fit the normalization
+            #       rules, rather than exploding
+            self.assertRaises(errors.InvalidNormalization,
+                    self.wt.smart_add, [])
+        finally:
+            osutils.normalized_filename = orig

=== modified file 'bzrlib/win32utils.py'
--- bzrlib/win32utils.py	2007-02-03 15:02:06 +0000
+++ bzrlib/win32utils.py	2007-07-02 22:36:11 +0000
@@ -238,3 +238,32 @@
 
 def get_host_name_unicode():
     return _ensure_unicode(get_host_name())
+
+
+def glob_expand_for_win32(file_list):
+    """Replacement for glob expansion by the shell.
+
+    Win32's cmd.exe does not do glob expansion (eg ``*.py``), so we do our own
+    here.
+
+    :param file_list: A list of filenames which may include shell globs.
+    :return: An expanded list of filenames.
+
+    Introduced in bzrlib 0.18.
+    """
+    if not file_list:
+        return []
+    import glob
+    expanded_file_list = []
+    for possible_glob in file_list:
+        glob_files = glob.glob(possible_glob)
+
+        if glob_files == []:
+            # special case to let the normal code path handle
+            # files that do not exists
+            expanded_file_list.append(possible_glob)
+        else:
+            expanded_file_list += glob_files
+    return expanded_file_list
+
+

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2007-07-04 01:53:26 +0000
+++ bzrlib/workingtree.py	2007-07-04 08:08:08 +0000
@@ -615,7 +615,7 @@
         # should probably put it back with the previous ID.
         # the read and write working inventory should not occur in this 
         # function - they should be part of lock_write and unlock.
-        inv = self.read_working_inventory()
+        inv = self.inventory
         for f, file_id, kind in zip(files, ids, kinds):
             assert kind is not None
             if file_id is None:
@@ -623,7 +623,7 @@
             else:
                 file_id = osutils.safe_file_id(file_id)
                 inv.add_path(f, kind=kind, file_id=file_id)
-        self._write_inventory(inv)
+            self._inventory_is_modified = True
 
     @needs_tree_write_lock
     def _gather_kinds(self, files, kinds):

